// SPDX-License-Identifier: MIT
// Copyright (c) 2021 KikooDX
// This file is part of [KBLE](https://sr.ht/~kikoodx/kble), which is
// MIT licensed. The MIT license requires this copyright notice to be
// included in all copies and substantial portions of the software.
const ray = @cImport({
    @cInclude("raylib.h");
});
const std = @import("std");
const maxInt = std.math.maxInt;
const log = std.log.info;

const conf = @import("conf.zig");
const Level = @import("level.zig");
const Vec2 = @import("vec2.zig");
const Rect = @import("rect.zig");
const Mouse = @import("mouse.zig");
const movement = @import("movement.zig");
const scaling = @import("scaling.zig");
const actions = @import("actions.zig");
const ActionsDef = actions.ActionsDef;
const ActionCat = actions.ActionCat;

const char_range = 255;

pub fn main() !void {
    // Create allocator
    var arena = std.heap.ArenaAllocator.init(std.heap.page_allocator);
    defer arena.deinit();
    const allocator = &arena.allocator;

    // Create window.
    ray.SetConfigFlags(ray.FLAG_WINDOW_RESIZABLE);
    ray.InitWindow(640, 480, "KBLE");
    defer ray.CloseWindow();

    // Limit FPS for performance.
    ray.SetTargetFPS(60);

    // Create level.
    var level: Level = try Level.init(allocator, 16, 16);
    defer level.deinit(allocator);

    // Create camera.
    var camera: Vec2 = Vec2.init(0, 0);

    // Init scale, used by drawing code.
    var scale: scaling.scale_type = scaling.scale_default;

    // Create mouse.
    var mouse: Mouse = Mouse.init();

    // Create cursor.
    var cursor: Vec2 = Vec2.init(0, 0);

    // Create binding "table".
    var bindings = [_]*const actions.Action{&actions.ActionsDef.none} ** char_range;
    // Set default bindings.
    // Movement.
    bindings[' '] = &ActionsDef.move_but_dont;
    bindings['h'] = &ActionsDef.move_left;
    bindings['H'] = &ActionsDef.move_LEFT;
    bindings['j'] = &ActionsDef.move_down;
    bindings['J'] = &ActionsDef.move_DOWN;
    bindings['k'] = &ActionsDef.move_up;
    bindings['K'] = &ActionsDef.move_UP;
    bindings['l'] = &ActionsDef.move_right;
    bindings['L'] = &ActionsDef.move_RIGHT;
    bindings['y'] = &ActionsDef.move_up_left;
    bindings['Y'] = &ActionsDef.move_UP_LEFT;
    bindings['u'] = &ActionsDef.move_up_right;
    bindings['U'] = &ActionsDef.move_UP_RIGHT;
    bindings['b'] = &ActionsDef.move_down_left;
    bindings['B'] = &ActionsDef.move_DOWN_LEFT;
    bindings['n'] = &ActionsDef.move_down_right;
    bindings['N'] = &ActionsDef.move_DOWN_RIGHT;
    // Verbs.
    bindings['d'] = &ActionsDef.verb_delete;
    bindings['r'] = &ActionsDef.verb_replace;
    // Scale.
    bindings['='] = &ActionsDef.scale_reset;
    bindings['+'] = &ActionsDef.scale_up;
    bindings['-'] = &ActionsDef.scale_down;

    // Create input buffer.
    const input_buffer_len = 255;
    var input_buffer: [input_buffer_len]u32 = undefined;
    comptime {
        var i: u8 = 0;
        while (i < input_buffer_len) : (i += 1) {
            input_buffer[i] = 0;
        }
    }
    var input_cursor: u8 = 0;

    while (!ray.WindowShouldClose()) {
        {
            // Get keyboard input.
            var key = ray.GetCharPressed();
            // Check if more characters have been pressed.
            while (key != 0) {
                // Add key to buffer.
                input_buffer[input_cursor] = @intCast(u32, key);
                input_cursor += 1;
                // Avoid writing out of memory.
                if (input_cursor >= input_buffer_len)
                    input_cursor = input_buffer_len - 1;

                key = ray.GetCharPressed();
            }
        }

        // Process buffer content.
        // Read the buffer backwards. This is placeholder logic.
        while (input_cursor > 0) {
            input_cursor -= 1;
            const key = if (input_buffer[input_cursor] <= char_range)
                input_buffer[input_cursor]
            else
                0;

            const action: actions.Action = bindings[key].*;

            switch (action.category) {
                .none => std.log.info("No action bound to {}.", .{key}),
                .movement => {
                    const selection_update: movement.SelectionUpdate =
                        action.function_move(&cursor, 1, action.toggle);
                    if (selection_update.active) {
                        level.apply_selection_update(selection_update);
                    }
                },
                .verb => {
                    action.function_verb(&level, 1);
                },
                .scale => {
                    scale = action.function_scale(scale);
                },
            }
        }

        // Mouse operations.
        if (conf.mouse_enabled) {
            // Update position.
            {
                // Set mouse offset.
                ray.SetMouseOffset(camera.x, camera.y);
                // Set mouse scaling.
                ray.SetMouseScale(1.0 / @intToFloat(f32, scale), 1.0 / @intToFloat(f32, scale));

                mouse.pos.x = @intCast(u16, ray.GetMouseX());
                mouse.pos.y = @intCast(u16, ray.GetMouseY());
            }

            const left_press: bool = ray.IsMouseButtonPressed(conf.mouse_left_btn);
            const left_release: bool = ray.IsMouseButtonReleased(conf.mouse_left_btn);
            const right_press: bool = ray.IsMouseButtonPressed(conf.mouse_right_btn);

            // State machine.
            switch (mouse.mode) {
                // See if can switch mode.
                .idle => {
                    const mod_rect_sel: bool = ray.IsKeyDown(ray.KEY_LEFT_SHIFT) or
                        ray.IsKeyDown(ray.KEY_RIGHT_SHIFT);
                    if (left_press) { // Switch mode.
                        cursor = mouse.pos;
                        mouse.start_pos = mouse.pos;
                        mouse.mode = if (mod_rect_sel)
                            Mouse.MouseMode.rect_sel
                        else
                            Mouse.MouseMode.sel;
                    }
                },
                // Select stuff under the cursor.
                .sel => {
                    cursor = mouse.pos;
                    level.select_cell(mouse.pos, true);
                },
                // Rectangle selection, see `left_release` for handling.
                .rect_sel => {},
            }

            // If left button is released, get out of current mode and apply changes if necessary.
            if (left_release) {
                defer mouse.mode = Mouse.MouseMode.idle;
                defer cursor = mouse.pos;

                // Select area.
                if (mouse.mode == Mouse.MouseMode.rect_sel) {
                    const selection = Rect.init_from_vec2(mouse.start_pos, mouse.pos);
                    level.select_rect(selection, true);
                }
            }

            // Reset selection on right click.
            if (right_press) {
                cursor = mouse.pos;
                level.apply_selection_update(movement.move_but_dont(&cursor, 1, true));
            }
        }

        ray.BeginDrawing();
        defer ray.EndDrawing();

        ray.ClearBackground(ray.BLACK);
        level.draw(scale, camera);
        level.draw_selection(scale, camera);
        //ray.DrawFPS(0, 0);
    }
}
